@@ -16,6 +16,10 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Author: Colin Ian King <colin.king@canonical.com>
+ *
+ * Note: Code is optimised to reduce CPU overhead. There are some arcane
+ * string scanning and string formatting specific functions to remove the
+ * overhead of the use of sscanf and sprintf.
  */
 #define _GNU_SOURCE
 
@@ -187,22 +191,22 @@ typedef struct {
 } scale_t;
 
 typedef struct {
-	uint32_t hash;
-	uint16_t offset;
+	uint32_t hash;			/* Hash of /proc/stat field tag */
+	uint16_t offset;		/* offset into proc_stat_t struct */
 } proc_stat_fields_t;
 
+/* ncurses mode or std tty mode display functions */
 typedef struct {
 	void (*df_setup)(void);
 	void (*df_endwin)(void);
 	void (*df_clear)(void);
 	void (*df_refresh)(void);
 	void (*df_winsize)(bool redo);
-	void (*df_printfnl)(const char *fmt, ...);
-	void (*df_putc)(int ch);
-	void (*df_putstr)(char *str, int n);
+	void (*df_putstrnl)(char *str, int n);
 	void (*df_linebreak)(void);
 } display_funcs_t;
 
+/* CPU frequency scale suffixes */
 static cpu_freq_scale_t cpu_freq_scale[] = {
 	{ 1e1,  1e0,  "Hz" },
 	{ 1e4,  1e3,  "KHz" },
@@ -213,6 +217,7 @@ static cpu_freq_scale_t cpu_freq_scale[] = {
 	{ -1.0, -1.0,  NULL }
 };
 
+/* seconds scale suffixes, secs, mins, hours, etc */
 static const scale_t second_scales[] = {
 	{ 's',	1 },
 	{ 'm',	60 },
@@ -252,6 +257,7 @@ static pid_t opt_pid = -1;		/* PID to match against, -p option */
 static bool resized;			/* window resized */
 static int rows = 25;			/* tty size, rows */
 static int cols = 80;			/* tty size, columns */
+static int cury = 0;			/* current y curpos */
 
 static display_funcs_t df;		/* display functions */
 
@@ -320,6 +326,131 @@ static void handle_sigwinch(int sig)
 	resized = true;
 }
 
+/*
+ *  duration_round()
+ *	round duration to nearest 1/100th second
+ */
+static inline double duration_round(const double duration)
+{
+        return floor((duration * 100.0) + 0.5) / 100.0;
+}
+
+/*
+ *  putint()
+ *	put a decimal value v into string str with max
+ *	length of nbytes
+ */
+static int OPTIMIZE3 HOT putint(char *str, int nbytes, int v, bool zeropad)
+{
+        register char *ptr = str + nbytes;
+	int ret = nbytes;
+	char pad;
+
+	*(ptr--) = '\0';
+	while (--nbytes >= 0) {
+		*(ptr--) = '0' + (v % 10);
+		v /= 10;
+		if (UNLIKELY(!v))
+			break;
+	}
+
+	pad = zeropad ? '0' : ' ';
+	while (--nbytes >= 0)
+		*(ptr--) = pad;
+
+	return ret;
+}
+
+
+/*
+ *  putuint()
+ *	put unsigned decimal value v into str
+ *      with no leading spaces.
+ */
+static int  OPTIMIZE3 HOT putuint(char *str, unsigned int v)
+{
+	register char *ptr = str;
+	register char *p1, *p2, *mid;
+
+	do {
+		*(ptr++) = '0' + (v % 10);
+		v /= 10;
+	} while (v);
+
+	*ptr = '\0';
+
+	/* and reverse */
+	mid = str + ((ptr - str) >> 1);
+	p1 = str;
+	p2 = ptr - 1;
+	while (p1 != mid) {
+		register char c = *p1;
+		*(p1++) = *p2;
+		*(p2--) = c;
+	}
+	return ptr - str;
+}
+
+/*
+ *  putdouble()
+ *	put a double in %6.2 with trailing space
+ */
+static int OPTIMIZE3 HOT putdouble(char *str, double v)
+{
+	v += 0.005;	/* Round up */
+
+	(void)putint(str, 3, (int)v, false);
+	str[3] = '.';
+	(void)putint(str + 4, 2, v * 100.0 - (double)((int)v * 100), true);
+	str[2] = (str[2] == ' ') ? '0' : str[2];
+	str[6] = ' ';
+	str[7] = '\0';
+
+	return 6;
+}
+
+/*
+ *  putdouble_decpl
+ *	put a double with decpl number of decimal places with trailing space, up to 4 decpl
+ */
+static int OPTIMIZE3 HOT putdouble_decpl(char *str, double v, int decpl)
+{
+	char *ptr = str;
+	static const double scales[] = {
+		1.0,
+		10.0,
+		100.0,
+		1000.0,
+		10000.0
+	};
+	double scale;
+
+	decpl = decpl > 4 ? 4 : decpl;
+	scale = scales[decpl];
+
+	v += 0.5 / scale;	/* Round up */
+	ptr += putuint(ptr, (int)v);
+	*(ptr++) = '.';
+	ptr += putint(ptr, decpl, scale * v - (double)((int)v * scale), true);
+	*ptr = '\0';
+
+	return ptr - str;
+}
+
+/*
+ *  putstr()
+ *	append a string and return bytes added
+ */
+static int OPTIMIZE3 HOT putstr(char *dst, int max, char *src)
+{
+	register int n = 0;
+
+	while (LIKELY(*(dst++) = *(src++)) && LIKELY(n < max))
+		n++;
+	*dst = '\0';
+	return n;
+}
+
 /*
  *   cpustat_top_setup
  *	setup display in top mode
@@ -343,7 +474,7 @@ static void cpustat_generic_winsize(bool redo)
 	if (redo) {
 		struct winsize ws;
 
-		if (ioctl(fileno(stdin), TIOCGWINSZ, &ws) != -1) {
+		if (LIKELY(ioctl(fileno(stdin), TIOCGWINSZ, &ws) != -1)) {
 			rows = ws.ws_row;
 			cols = ws.ws_col;
 		} else {
@@ -403,47 +534,16 @@ static inline void cpustat_top_refresh(void)
 }
 
 /*
- * cpustat_generic_printfnl()
- *      cpustat generic printf with newline
+ *  cpustat_top_putstrnl()
+ * 	cpustat put string in top mode with newline
+ *	(or not if there is potential for line wrap)
  */
-static void cpustat_generic_printfnl(const char *fmt, ...)
+static void cpustat_top_putstrnl(char *str, int n)
 {
-	va_list ap;
-	int n;
+	if (UNLIKELY(cury >= rows))
+		return;
 
-	va_start(ap, fmt);
-	char buf[256];
-
-	n = vsnprintf(buf, sizeof(buf) - 2, fmt, ap);
-	df.df_putstr(buf, n);
-	va_end(ap);
-}
-
-/*
- *  cpustat_top_putc()
- *	cpustat put char in top mode
- */
-static void cpustat_top_putc(int ch)
-{
-	addch(ch);
-}
-
-/*
- *  cpustat_nomnal_putc()
- *	cpustat put char in normal mode
- */
-static void cpustat_normal_putc(int ch)
-{
-	fputc(ch, stdout);
-}
-
-/*
- *  cpustat_top_putstr)
- * 	cpustat put string in top mode
- */
-static void cpustat_top_putstr(char *str, int n)
-{
-	if (n > cols) {
+	if (UNLIKELY(n > cols)) {
 		n = cols;
 		str[n] = '\0';
 	} else {
@@ -451,16 +551,17 @@ static void cpustat_top_putstr(char *str, int n)
 		str[n + 1] = '\0';
 	}
 
+	cury++;
 	addstr(str);
 }
 
 /*
- *  cpustat_normal_putstr()
- * 	cpustat put string in normal mode
+ *  cpustat_normal_putstr(()
+ * 	cpustat put string in normal mode with newline
  */
-static void cpustat_normal_putstr(char *str, int n)
+static void cpustat_normal_putstrnl(char *str, int n)
 {
-	if (n > cols)
+	if (UNLIKELY(n > cols))
 		n = cols;
 
 	str[n] = '\n';
@@ -478,27 +579,25 @@ static void cpustat_normal_linebreak(void)
 	putc('\n', stdout);
 }
 
+/* 'top' mode display functions */
 static display_funcs_t df_top = {
 	cpustat_top_setup,
 	cpustat_top_endwin,
 	cpustat_top_clear,
 	cpustat_top_refresh,
 	cpustat_top_winsize,
-	cpustat_generic_printfnl,
-	cpustat_top_putc,
-	cpustat_top_putstr,
+	cpustat_top_putstrnl,
 	cpustat_noop,
 };
 
+/* normal tty mode display functions */
 static display_funcs_t df_normal = {
 	cpustat_noop,
 	cpustat_noop,
 	cpustat_noop,
 	cpustat_noop,
 	cpustat_generic_winsize,
-	cpustat_generic_printfnl,
-	cpustat_normal_putc,
-	cpustat_normal_putstr,
+	cpustat_normal_putstrnl,
 	cpustat_normal_linebreak,
 };
 
@@ -514,7 +613,7 @@ static uint64_t OPTIMIZE3 HOT strtouint64(char *str, char **endptr)
 	for (;;) {
 		register unsigned int digit = *str - '0';
 
-		if (digit > 9)
+		if (UNLIKELY(digit > 9))
 			break;
 		if (UNLIKELY(v >= 1844674407370955161ULL))
 			goto do_overflow;
@@ -542,7 +641,7 @@ static uint32_t OPTIMIZE3 HOT strtouint32(char *str, char **endptr)
 	for (;;) {
 		register unsigned int digit = *str - '0';
 
-		if (digit > 9)
+		if (UNLIKELY(digit > 9))
 			break;
 		if (UNLIKELY(v >= 429496729))
 			goto do_overflow;
@@ -562,7 +661,7 @@ do_overflow:
  *  get_ticks()
  *	get ticks
  */
-static inline uint64_t get_ticks(void)
+static inline uint64_t OPTIMIZE3 HOT get_ticks(void)
 {
 	return (opt_flags & OPT_TICKS_ALL) ?
 		clock_ticks * (uint64_t)sysconf(_SC_NPROCESSORS_ONLN) :
@@ -573,7 +672,7 @@ static inline uint64_t get_ticks(void)
  *  secs_to_str()
  *	report seconds in different units.
  */
-static const char *secs_to_str(const double secs)
+static char *secs_to_str(const double secs)
 {
 	static char buf[16];
 	int i;
@@ -582,8 +681,9 @@ static const char *secs_to_str(const double secs)
 		if (secs <= second_scales[i + 1].scale)
 			break;
 	}
-	snprintf(buf, sizeof(buf), "%5.2f%c",
-		secs / second_scales[i].scale, second_scales[i].ch);
+	putdouble(buf, secs / second_scales[i].scale);
+	buf[6] = second_scales[i].ch;
+
 	return buf;
 }
 
@@ -652,9 +752,20 @@ static void handle_sig(int dummy)
 
 /*
  *  count_bits()
- *	count bits set, from C Programming Language 2nd Ed
  */
-static unsigned int count_bits(const unsigned int val)
+#if defined(__GNUC__)
+/*
+ *  use GCC built-in
+ */
+static inline unsigned int count_bits(const unsigned int val)
+{
+	return __builtin_popcount(val);
+}
+#else
+/*
+ *  count bits set, from C Programming Language 2nd Ed
+ */
+static inline unsigned int OPTIMIZE3 HOT count_bits(const unsigned int val)
 {
 	register unsigned int c, n = val;
 
@@ -663,6 +774,7 @@ static unsigned int count_bits(const unsigned int val)
 
 	return c;
 }
+#endif
 
 /*
  *  get_pid_cmdline
@@ -671,7 +783,7 @@ static unsigned int count_bits(const unsigned int val)
 static char *get_pid_cmdline(const pid_t pid)
 {
 	static char buffer[4096];
-	char *ptr = NULL;
+	char *ptr;
 	int fd;
 	ssize_t ret;
 	pid_info_t *info;
@@ -680,7 +792,12 @@ static char *get_pid_cmdline(const pid_t pid)
 	struct stat statbuf;
 	bool statok = false;
 
-	snprintf(path, sizeof(path), "/proc/%i/cmdline", pid);
+	ptr = path;
+	ptr += putstr(ptr, 6, "/proc/");
+	ptr += putuint(ptr, pid);
+	putstr(ptr, 8, "/cmdline");
+	ptr = NULL;
+
 	if (UNLIKELY((fd = open(path, O_RDONLY)) < 0))
 		return NULL;
 
@@ -891,55 +1008,33 @@ static int info_compare_total(const void *const item1, const void *const item2)
 	return ((*info2)->total > (*info1)->total) ? 1 : -1;
 }
 
-/*
- *  duration_round()
- *	round duration to nearest 1/100th second
- */
-static inline double duration_round(const double duration)
-{
-        return floor((duration * 100.0) + 0.5) / 100.0;
-}
-
-static inline void putdec(char *str, const int v)
-{
-	register int d;
-
-	d = (v / 10);
-	str[0] = d > 9 ? '?' : '0' + d;
-	d = v % 10;
-	str[1] = '0' + d;
-}
-
 /*
  *  info_banner_dump()
  *	dump banner for per_info stats
  */
 static void info_banner_dump(const double time_now)
 {
-	char str[256] = "  %CPU   %USR   %SYS   PID S  CPU   Time Task";
+	char str[256] = "  %CPU   %USR   %SYS   PID S  CPU    Time Task";
 	char *ptr = str + strlen(str);
 
-	if (opt_flags & OPT_TIMESTAMP) {
+	if (UNLIKELY(opt_flags & OPT_TIMESTAMP)) {
 		struct tm tm;
 
 		get_tm(time_now, &tm);
 		strncpy(ptr, "  (", 3);
 		ptr += 3;
-		putdec(ptr, tm.tm_hour);
-		ptr += 2;
+		ptr += putint(ptr, 2, tm.tm_hour, false);
 		*ptr = ':';
 		ptr++;
-		putdec(ptr, tm.tm_min);
-		ptr += 2;
+		ptr += putint(ptr, 2, tm.tm_min, false);
 		*ptr = ':';
 		ptr++;
-		putdec(ptr, tm.tm_sec);
-		ptr += 2;
+		ptr += putint(ptr, 2, tm.tm_sec, false);
 		*ptr = ')';
 		ptr++;
 	}
 	*ptr = '\0';
-	df.df_putstr(str, ptr - str);
+	df.df_putstrnl(str, ptr - str);
 }
 
 /*
@@ -954,6 +1049,8 @@ static void info_dump(
 	double *u_total,
 	double *s_total)
 {
+	char buffer[512], *ptr = buffer;
+
 	const double cpu_u_usage =
 		total_ticks == 0 ? 0.0 : 100.0 * (double)uticks / total_ticks;
 	const double cpu_s_usage =
@@ -963,31 +1060,50 @@ static void info_dump(
 	*u_total += cpu_u_usage;
 	*s_total += cpu_s_usage;
 
-	df.df_printfnl("%6.2f %6.2f %6.2f %5d %c %4d %s %s%s%s",
-		cpu_u_usage + cpu_s_usage,
-		cpu_u_usage, cpu_s_usage,
-		info->pid,
-		info->state,
-		info->processor,
-		secs_to_str(cpu_time),
-		info->kernel_thread ?
-			"[" : "",
-		info->cmdline,
-		info->kernel_thread ?
-			"]" : "");
+	ptr += putdouble(ptr, cpu_u_usage + cpu_s_usage);
+	*(ptr++) = ' ';
+	ptr += putdouble(ptr, cpu_u_usage);
+	*(ptr++) = ' ';
+	ptr += putdouble(ptr, cpu_s_usage);
+	*(ptr++) = ' ';
+	ptr += putint(ptr, 5, info->pid, false);
+	*(ptr++) = ' ';
+	*(ptr++) = info->state;
+	*(ptr++) = ' ';
+	ptr += putint(ptr, 4, info->processor, false);
+	*(ptr++) = ' ';
+	ptr += putstr(ptr, 20, secs_to_str(cpu_time));
+	*(ptr++) = ' ';
+	if (info->kernel_thread)
+		*(ptr++) = '[';
+	ptr += putstr(ptr, 128, info->cmdline);
+	if (info->kernel_thread)
+		*(ptr++) = ']';
+	*ptr = '\0';
+
+	df.df_putstrnl(buffer, ptr - buffer);
 }
 
 /*
  *  info_total_dump()
  *	dump out totals of total, system and user times
  */
-static void info_total_dump(
+static inline void info_total_dump(
 	const double u_total,
 	const double s_total)
 {
-	if (opt_flags & OPT_TOTAL)
-		df.df_printfnl("%6.2f %6.2f %6.2f Total",
-			u_total + s_total, u_total, s_total);
+	if (UNLIKELY(opt_flags & OPT_TOTAL)) {
+		char buffer[256], *ptr = buffer;
+
+		ptr += putdouble(ptr, u_total + s_total);
+		*(ptr++) = ' ';
+		ptr += putdouble(ptr, u_total);
+		*(ptr++) = ' ';
+		ptr += putdouble(ptr, s_total);
+		*(ptr++) = ' ';
+		ptr += putstr(ptr, 5, "Total");
+		df.df_putstrnl(buffer, ptr - buffer);
+	}
 }
 
 /*
@@ -1028,7 +1144,7 @@ static void samples_dump(
 	if (opt_flags & OPT_GRAND_TOTAL) {
 		double cpu_u_total = 0.0, cpu_s_total = 0.0;
 
-		df.df_printfnl("Grand Total (from %" PRIu32 " samples, %.1f seconds):",
+		printf("Grand Total (from %" PRIu32 " samples, %.1f seconds):\n",
 			samples, duration);
 		info_banner_dump(time_now);
 		for (i = 0; i < n; i++) {
@@ -1038,7 +1154,7 @@ static void samples_dump(
 				&cpu_u_total, &cpu_s_total);
 		}
 		info_total_dump(cpu_u_total, cpu_s_total);
-		df.df_printfnl("");
+		putchar('\n');
 	}
 
 	if (!filename) {
@@ -1159,7 +1275,7 @@ static void samples_distribution(const uint64_t nr_ticks)
 	sample_delta_list_t *sdl;
 	unsigned int bucket[MAX_DIVISIONS], max_bucket = 0, valid = 0, i, total = 0;
 	double min = DBL_MAX, max = -DBL_MAX, division, prev;
-	double scale = 100.0 / (double)nr_ticks;
+	const double scale = 100.0 / (double)nr_ticks;
 
 	memset(bucket, 0, sizeof(bucket));
 
@@ -1579,15 +1695,22 @@ static inline void proc_stat_dump(
 	const proc_stat_t *delta,
 	const double duration)
 {
-	double scale = 1.0 / duration;
-	df.df_printfnl("%.1f Ctxt/s, %.1f IRQ/s, %.1f softIRQ/s, "
-		"%.1f new tasks/s, %" PRIu64 " running, %" PRIu64 " blocked\n",
-		scale * delta->ctxt,
-		scale * delta->irq,
-		scale * delta->softirq,
-		scale * delta->processes,
-		delta->running,
-		delta->blocked);
+	const double scale = 1.0 / duration;
+	char buffer[128], *ptr = buffer;
+
+	ptr += putdouble_decpl(ptr, scale * delta->ctxt, 1);
+	ptr += putstr(ptr, 9, " Ctxt/s, ");
+	ptr += putdouble_decpl(ptr, scale * delta->irq, 1);
+	ptr += putstr(ptr, 8, " IRQ/s, ");
+	ptr += putdouble_decpl(ptr, scale * delta->softirq, 1);
+	ptr += putstr(ptr, 12, " softIRQ/s, ");
+	ptr += putdouble_decpl(ptr, scale * delta->processes, 1);
+	ptr += putstr(ptr, 14, " new tasks/s, ");
+	ptr += putuint(ptr, delta->running);
+	ptr += putstr(ptr, 10, " running, ");
+	ptr += putuint(ptr, delta->blocked);
+	ptr += putstr(ptr, 9, " blocked");
+	df.df_putstrnl(buffer, ptr - buffer);
 }
 
 /*
@@ -1614,7 +1737,7 @@ static void get_cpustats(
 		cpu_info_t info;
 		uint64_t utime;
 		uint64_t stime;
-		char filename[PATH_MAX];
+		char filename[PATH_MAX], *fnptr;
 		char buffer[4096];
 		char *ptr = buffer, *endptr, *tmp;
 		ssize_t len;
@@ -1623,8 +1746,10 @@ static void get_cpustats(
 		if (!isdigit(entry->d_name[0]))
 			continue;
 
-		snprintf(filename, sizeof(filename), "/proc/%s/stat",
-			entry->d_name);
+		fnptr = filename;
+		fnptr += putstr(fnptr, 6, "/proc/");
+		fnptr += putstr(fnptr, PATH_MAX - 6, entry->d_name);
+		fnptr += putstr(fnptr, 5, "/stat");
 		if ((fd = open(filename, O_RDONLY)) < 0)
 			continue;
 
@@ -1721,17 +1846,19 @@ static void get_cpustats(
  */
 static double cpu_freq_average(uint32_t max_cpus)
 {
-	uint32_t i, n = 0;
+	unsigned int i, n = 0;
 	double total_freq = 0;
 
 	for (i = 0; i < max_cpus; i++) {
-		char path[PATH_MAX];
+		char filename[PATH_MAX], *fnptr;
 		int fd;
 
-		snprintf(path, sizeof(path),
-			"/sys/devices/system/cpu/cpu%" PRIu32
-			"/cpufreq/scaling_cur_freq", i);
-		if (LIKELY((fd = open(path, O_RDONLY)) > -1)) {
+		fnptr = filename;
+		fnptr += putstr(fnptr, 28, "/sys/devices/system/cpu/cpu");
+		fnptr += putuint(fnptr, i);
+		fnptr += putstr(fnptr, 25, "/cpufreq/scaling_cur_freq");
+
+		if (LIKELY((fd = open(filename, O_RDONLY)) > -1)) {
 			char buffer[64];
 			ssize_t ret;
 
@@ -1755,10 +1882,10 @@ static double cpu_freq_average(uint32_t max_cpus)
  *  cpu_freq_format()
  *	scale cpu freq into a human readable form
  */
-static const char *cpu_freq_format(double freq)
+static char *cpu_freq_format(double freq)
 {
 	static char buffer[40];
-	char *suffix = "EHz";
+	char *suffix = "EHz", *ptr = buffer;
 	double scale = 1e18;
 	size_t i;
 
@@ -1770,8 +1897,9 @@ static const char *cpu_freq_format(double freq)
 		}
 	}
 
-	snprintf(buffer, sizeof(buffer), "%.2f %s",
-		freq / scale, suffix);
+	ptr += putdouble_decpl(buffer, freq / scale, 2);
+	*(ptr++) = ' ';
+	ptr += putstr(ptr, 3, suffix);
 
 	return buffer;
 }
@@ -1874,6 +2002,26 @@ unknown:
 
 }
 
+/*
+ *  load_online_dump()
+ *	dump load and cpu related stats
+ */
+static inline void load_online_dump(const uint32_t max_cpus)
+{
+	double avg_cpu_freq = cpu_freq_average(max_cpus);
+	char buffer[128], *ptr = buffer;
+
+	ptr += putstr(ptr, 10, "Load Avg ");
+	ptr += putstr(ptr, 20, load_average());
+	ptr += putstr(ptr, 12, ", Freq Avg ");
+	ptr += putstr(ptr, 20, cpu_freq_format(avg_cpu_freq));
+	ptr += putstr(ptr, 2, ", ");
+	ptr += putstr(ptr, 10, cpus_online());
+	ptr += putstr(ptr, 12, " CPUs online");
+
+	df.df_putstrnl(buffer, ptr - buffer);
+}
+
 /*
  *  show_usage()
  *	show how to use
@@ -2082,6 +2230,7 @@ int main(int argc, char **argv)
 		double secs, duration, right_now;
 
 		df.df_clear();
+		cury = 0;
 
 		/* Timeout to wait for in the future for this sample */
 		secs = time_start + ((double)t * duration_secs) - time_now;
@@ -2132,15 +2281,10 @@ retry:
 		df.df_refresh();
 
 		if (opt_flags & OPT_EXTRA_STATS) {
-			double avg_cpu_freq = cpu_freq_average(max_cpus);
-
 			get_proc_stat(proc_stat_new);
 			proc_stat_diff(proc_stat_old, proc_stat_new,
 					&proc_stat_delta);
-			df.df_printfnl("Load Avg %s, Freq Avg. %s, %s CPUs online",
-				load_average(),
-				cpu_freq_format(avg_cpu_freq),
-				cpus_online());
+			load_online_dump(max_cpus);
 			proc_stat_dump(&proc_stat_delta, duration);
 		}
 
